/*
 * Cremetz Olivier
 * Benouhamane el Medhi
 *
 * Tp de Graphe : Plus court chemin dans le metro parisien
 * A rendre le 01-04-2005
 *
 */


#include <stdio.h> /* printf, scanf */
#include <stdlib.h> /* pour malloc */
#include <string.h> /* pour strcpy */


/* Constantes booleennes */
#define VRAI 1
#define FAUX 0

/* Constante infini : Ici on met l'infini a 100000
   On ne atteindra pas cette valeur ici */
#define INFINITY 100000
#define DEP 1
#define ARRIV 2


/* Definition du type booleen */
typedef int booleen;



/* Structure definissant la liste des successeurs d'un sommet */
typedef struct Arcs{
  int numStation;
  float poids;
  struct Arcs *suivant;
} Arcs, *pt_Arcs;



/* Structure definissant un sommet
 * Un sommet est compose de :
 * son nom de station
 * son potentiel (calcule dans dijkstra)
 * sa marque (si on est deja passe par ce sommet avec un chemin minimal)
 * la station mere (le sommet precedent qui marque ce sommet)
 * ses coordonnees
 * sa liste de ses successeurs
 */
typedef struct Sommets{
  char * nomStation;
  float potentiel;
  booleen marque;
  int stationMere;
  int coordX;
  int coordY;
  pt_Arcs successeurs;
} Sommets,*pt_Sommets;




/* Liste (des lignes) des Stations*/
typedef struct Station{
  int numStation;
  struct Station *suivant;
}Station,*pt_stations; 


/* Liste (des lignes) des Stations de Depart */
pt_stations teteDepart=NULL;

/* Liste (des lignes) des Stations d'arrivee */
pt_stations teteArrivee=NULL;



/* Liste bidirectionnelle contenant le plus court chemin trouve */
typedef struct cheminFinal{
  int numStation;
  float potentiel;
  int stationMere;
  struct cheminFinal *suivant;
  struct cheminFinal *precedent;
} cheminFinal, *pt_listeChemins;

pt_listeChemins tete=NULL;

/* Nombre de sommets marques */
int nbSommetsMarques;






/**************************************************************************
 * Procedure get_nom(char **nom,FILE *f)
 *
 * Donnee(s) :  f: pointeur sur FILE :  handle du fichier
 * Donnee(s)/Resultat(s) : 
 * Resultat(s) : nom : pointeur sur une chaine de caractere
 *
 * Cette fonction lit le reste d'une ligne d'un fichier et revoit la chaine 
 * lue
 **************************************************************************/



void get_nom(char **nom,FILE *f)
{
  char temp[100];
  /* on lit dans le fichier jusqu'a \n */
  fgets(temp,100,f);
  
  /* On alloue un tableau de caractere de la taille de tmp + \0 */ 
  (*nom) = (char *)malloc((strlen(temp)+1)*sizeof(char));
  
  /* On termine la chaine par \0 au lieu de \n */
  temp[strlen(temp)-1]='\0';
  
  /* On recopie la chaine lue dans la chaine de retour*/
  strcpy(*nom, temp);
}




/**************************************************************************
 * procedure chargement_Fichier(pt_Sommets Graphe[],int n, int m, FILE *f)
 *
 * Donnee(s) : n : entier nombre de sommets
 *             m : entier nombre d'arcs
 *             f : pointeur sur handle du fichier
 * Resultat(s) : pt_Sommets Graphe[] : tableau de sommet du graphe de taille n
 *
 * Cette procedure lit le fichier passe en parametre du fichier et cree
 * le graphe en remplissant le tableau Graphe
 **************************************************************************/

void chargement_Fichier(pt_Sommets Graphe[],int n, int m, FILE *f)
{
  /* indices de boucle */
  int i;
  
  /* Variables pour les commentaires */
  int tmp;
  char *commentaire;
  
  /* Cellules temporaires */  
  pt_Sommets tmpSom;
  pt_Arcs tmpSucc;
  
  /* coordonnees */
  int coordX;
  int coordY;
  
  /* definition d'un arc */
  int SommetInit;
  int SommetFinal; 
  float poids;
  
  /*Lecture de la ligne de commentaire */
  get_nom(&commentaire,f);
  
  
  /* Recuperation du nom de la station dans la structure */
  printf("Recuperation des noms des stations dans la structure \n");
  for (i=0;i<n;i++)
    {
      fscanf(f,"%d ",&tmp);
      
      /* creation de la cellule Sommets */
      tmpSom=(pt_Sommets)malloc(sizeof(Sommets));
      get_nom(&tmpSom->nomStation,f);
      tmpSom->successeurs=NULL;
      
      /* On relie cette cellule au graphe */
      Graphe[tmp]=tmpSom;
    }
  
  /* Lecture de la ligne de commentaire */
  get_nom(&commentaire,f);
  
  /* recuperation des coordonnees inutiles dans ce programme */
  printf("Recuperation des coordonnees \n");
  for (i=0;i<n;i++)
    {
      fscanf(f,"%d %d %d\n",&tmp,&coordX,&coordY);
      
      Graphe[tmp]->coordX=coordX;
      Graphe[tmp]->coordY=coordY;
    }
  
  
  /* Lecture de la ligne de commentaire */
  get_nom(&commentaire,f);
  
  printf("Recuperation des %d arcs \n",m);
  for (i=0;i<m;i++)
    {
      fscanf(f,"%d %d %f",&SommetInit,&SommetFinal,&poids);
      
      /* Creation d'une cellule Arcs */
      tmpSucc=(pt_Arcs)malloc(sizeof(Arcs));
      tmpSucc->numStation=SommetFinal;
      tmpSucc->poids=poids;
      
      /* On relie cette nouvelle cellule au graphe */
      tmpSucc->suivant=Graphe[SommetInit]->successeurs;
      Graphe[SommetInit]->successeurs=tmpSucc;
    }
}


/**************************************************************************
 * procedure help_Station(pt_Sommets Graphe[],int n)
 *
 * Donnee(s) : Graphe : tableau de Sommets representant le graphe
 *             n : nombre de sommets
 * Donnee(s)/resultat(s) : 
 * Resultat(s) :
 *
 * Cette procedure affiche la liste des stations page par page (sur 22 lignes)
 **************************************************************************/

void help_Station(pt_Sommets Graphe[],int n)
{
  int i;
  
  for(i=0;i<n;i++)
    {
      printf("%4d %s\n",i,Graphe[i]->nomStation);
      
      if (i!=0 && i%22==0){
	printf("Appuyez sur entree pour continuer");
	getchar();
      }
    }
}




/**************************************************************************
 * fonction : ajouterStationFin(teteStation,precedent,numstation,
 * departOuArrivee) : retoure pointeur sur station pt_stations
 *
 * Donnee(s) : teteStation : pointeur sur la tete de liste de Stations
 *             precedent : pointeur sur la cellule precedente
 *             numstation : numero de Station a ajouter
 *             departouArrivee : entier sert a determiner si il s'agit de
 *             la liste de depart ou d'arriver : Cela permet de ne pas ecrire
 *             2 fonctions differentes
 *
 * Cette fonction ajoute une station a la liste (de depart ou d'arrivee)
 **************************************************************************/

pt_stations ajouterStationFin(pt_stations teteStation,pt_stations precedent,int numstation, int departOuArrivee)
{
  /* Creation d'une cellule */
  pt_stations tmp;
  tmp=(pt_stations)malloc(sizeof(Station));
  tmp->numStation=numstation;
  tmp->suivant=NULL;
  
  if (precedent!=NULL)
    {
      /* on ajoute en queue */
      precedent->suivant=tmp;
    }
  else
    /* On ajoute en tete*/ 
    teteStation=tmp;
  
  /* Modification des variables globales */
  if (departOuArrivee==DEP)
    teteDepart=teteStation;
  if (departOuArrivee==ARRIV)
    teteArrivee=teteStation;
  
  return tmp;
}




/**************************************************************************
 * procedure AfficherListe(pt_stations liste)
 *
 * Donnee(s) : liste : pointeur sur tete de liste des Stations
 * Donnee(s)/Resultat(s) :
 * Resultat(s) :
 *
 * Cette procedure affiche la liste des stations (de depart ou d'arrivee)
 **************************************************************************/
void AfficherListe(pt_stations liste)
{
  while (liste != NULL)
    {
      printf(" %d ,",liste->numStation);
      liste=liste->suivant;
    }
}



/**************************************************************************
 * fonction estDansListe(int station,pt_stations liste) : booleen
 *
 * Donnee(s) : station : entier numero de station
 *             liste : pointeur pt_stations sur le debut de la liste
 *
 * Verification si la station est dans la liste de station
 **************************************************************************/
booleen estDansListe(int station,pt_stations liste)
{
  while (liste != NULL)
    {
      if(station==liste->numStation)
	{
	  return VRAI;
	}
      liste=liste->suivant;
    }
  return FAUX;
}



/**************************************************************************
 * procedure rechercheStations(pt_Sommets Graphe[],int n, int departOuArrivee)
 *
 * Donnee(s) : Graphe : tableau de Sommets representant le graphe
 *             n : nombre de sommets
 *             departOuArrivee : entier sert a determiner si il s'agit de
 *             la liste de depart ou d'arriver : Cela permet de ne pas ecrire
 *             2 fonctions differentes
 * Donnee(s)/Resultat(s) :
 * Resultat(s) :
 *
 * Cette procedure demande de saisir un nom de station et ajoute toutes
 * les lignes a la liste de station(de depart ou d'arrivee) 
 * Ici on compare les chaines sans se soucier de la casse 
 **************************************************************************/
void rechercheStations(pt_Sommets Graphe[],int n, int departOuArrivee)
{
  int i;
  char * stationSouhaitee;
  
  pt_stations depart=teteDepart;
  pt_stations arrivee=teteArrivee;
  
  printf("Entrer le nom de la station (h pour l'aide):");
  get_nom(&stationSouhaitee,stdin);
  
  
  if(strcasecmp(stationSouhaitee,"h")==0)
    {
      /* Appel de l'aide */
      help_Station(Graphe,n);
      
      printf("Entrer la station :");
      get_nom(&stationSouhaitee,stdin);
    }
  
  printf("La station souhaitee est %s\n",stationSouhaitee);
  
  for (i=0;i<n;i++)
    {
      if(strcasecmp(Graphe[i]->nomStation,stationSouhaitee)==0)
	{
	  /* Ajout dans la liste des numero de la station trouvee */

	  if (departOuArrivee==DEP)
	    {
	      depart=ajouterStationFin(teteDepart,depart,i,DEP);
	    }
	  
	  if (departOuArrivee==ARRIV)
	    {
	      arrivee=ajouterStationFin(teteArrivee,arrivee,i,ARRIV);
	    }
	}
    }
}


/**************************************************************************
 * fonction ajouterFin(precedent,numstation, potentiel, stationMere) :
 * retourne un pointeur sur la nouvelle cellule cree
 * 
 * Donnee(s) : precedent : pointeur sur liste representant le chemin Final 
 *             numstation : entier numero de station marquee a ajouter
 *             potentiel : reel potentiel de la station
 *             stationMere : entier numero de la station ayant marque celle ci
 *
 * Cette procedure ajoute en fin de liste une cellule dans la liste 
 * bidirectionnelle representant le chemin final
 **************************************************************************/
pt_listeChemins ajouterFin(pt_listeChemins precedent,int numstation, float potentiel, int stationMere)
{
  
  pt_listeChemins tmp;
  tmp=(pt_listeChemins)malloc(sizeof(cheminFinal));
  tmp->numStation=numstation;
  tmp->potentiel=potentiel;
  tmp->stationMere=stationMere;
  tmp->suivant=NULL;
  /* On relie le precedent*/ 
  tmp->precedent=precedent;
  
  if (precedent!=NULL)
    {
      /* on ajoute en fin*/
      precedent->suivant=tmp;
    }
  else
    tete=tmp;
  return tmp;
}



/**************************************************************************
 * procedure afficherResultat(pt_listeChemins tete,pt_Sommets Graphe[])
 *
 * Donnee(s) : tete : pointeur sur la tete de la liste du chemin final
 *             Graphe : tableau de Sommets representant le graphe
 * Donnee(s)/Resultat(s) :
 * Resultat(s) :
 *
 * Cette procedure affiche l'itineraire avec les differentes etapes
 * ainsi que la valuation de l'arc
 **************************************************************************/
void afficherResultat(pt_listeChemins tete,pt_Sommets Graphe[])
{
  float longueurChemin=0;
  pt_listeChemins ptr;
  ptr=tete;
  int i=0;
  if (ptr!=NULL)
    {
      while (ptr!=NULL)
	{
	  printf("%3d) %s (Station %d) : %.2f\n", i, Graphe[ptr->numStation]->nomStation,ptr->numStation,ptr->potentiel);
	  i++;
	  longueurChemin=ptr->potentiel;
	  ptr=ptr->suivant;
	}
      printf("Le chemin a pour longeur : %.2f \n",longueurChemin);
    }
}







/**************************************************************************
 * fonction minimum(pt_Sommets Graphe[],int n) : retourne un entier
 *
 * Donnee(s) : Graphe : tableau de Sommets representant le graphe
 *             n : nombre de sommets
 * Donnee(s)/resultat(s) : 
 * Resultat(s) :
 *
 * Cette fonction recherche dans le tableau de Graphe le minimum des 
 * potentiels des sommets non marques
 * Cette fonction gere aussi une priorite en cas d'egalite de minimum
 * Elle prend en compte si il s'agit d'un sommet final ou pas (ce qui
 * reduit les calculs pour dijkstra)
 **************************************************************************/

int minimum(pt_Sommets Graphe[],int n)
{
  int i;
  int j;
  float mini;
  pt_stations arrivee=teteArrivee;
  /* x n'appartenant pas a S dont son potentiel est minimal */
  
  j=0;
  /* On recherche le premier sommet non marque */
  while (Graphe[j]->marque!=FAUX && j<n)
    {
      j++; 
    }
  
  if (j<n)
    {
      /* initialisation du minimum */
      mini=Graphe[j]->potentiel;
    } 
  else
    {
      /* Erreur pas de minimum trouve */
      fprintf(stderr,"Erreur dans le calcul du minimum\n");
      exit (6);
    }
  
  for (i=0;i<n;i++)
    {
      if(Graphe[i]->marque==FAUX)
	{
	  /*Gestion de la priorite */
	  if (Graphe[i]->potentiel==mini && estDansListe(i,arrivee))
	    {
	      mini=Graphe[i]->potentiel;
	      j=i;
	    }
	  
	  if (Graphe[i]->potentiel<mini)
	    {
	      /* On change le potentiel*/
	      mini=Graphe[i]->potentiel;
	      j=i;
	    }
	}
    }
  
  /* On retourne le minimum */  
  return j; 
}





/**************************************************************************
 * procedure genereChemin(pt_listeChemins finListe)
 *
 * Donnee(s) : finListe : pointeur sur la fin de la liste du chemin final
 * Donnee(s)/Resultat(s) :
 * Resultat(s) :
 *
 * La procedure dijkstra revoie l'arborescence a partir du sommet de depart 
 * jusqu'a l'arrivee. Cette procedure contient donc des arcs inutiles pour
 * montrer notre itineraire
 * Ici on passe la fin de liste en parametre car la liste est bidirectionnelle
 * et cela nous evite d'aller jusqu'a la fin pour rien
 * Cette procedure sera lancee a la fin de dijkstra
 * On regarde les precedents de tous les sommets marques et on ne garde que
 * ceux qui forment une chaine entre le sommet de fin et celui du debut
 **************************************************************************/
void genereChemin(pt_listeChemins finListe)
{
  /* on a marque le sommet final */
  nbSommetsMarques=1;
  pt_listeChemins tmp;
  int numStationPrecedente;
  numStationPrecedente = finListe->stationMere;
  finListe=finListe->precedent;
  printf("\n");
  while (finListe != NULL)
    {
      if (finListe->numStation == numStationPrecedente)
	{
	  /* On garde le sommet car il fait parti de notre chemin */
	  numStationPrecedente = finListe->stationMere;
	  nbSommetsMarques++;
	}
      
      else
	{
	  tmp=finListe;
	  /* On supprime la cellule */ 
	  if (finListe->precedent != NULL)
	    {
	      /* suppression dans la liste hors tete*/
	      (finListe->precedent)->suivant=tmp->suivant;
	      tmp->suivant->precedent=tmp->precedent;
	    }
	  else
	    {
	      /* suppression en tete */
	      tete=finListe->suivant;
	    }
	  /* On libere la memoire de la cellule */
	  free(tmp);
	}
      finListe=finListe->precedent;
    }
}



/**************************************************************************
 * procedure dijkstra(pt_Sommets Graphe[],int n)
 * 
 * Donnee(s) : Graphe : tableau de Sommets representant le graphe
 *             n : nombre de sommets
 * Donnee(s)/resultat(s) : 
 * Resultat(s) :
 *
 * Cette procedure cree l'arborescence des plus court chemin
 * a partir d'un sommet racine
 * On calcule les plus courtes distances de proche en proche par ajustements 
 * successifs 
 * l'algorithme fournit l'arboresence des plus court chemin qui sera tout 
 * de suite traitee pour trouver le chemin le plus court
 **************************************************************************/


void dijkstra(pt_Sommets Graphe[],int n)
{
  int i;
  int k;
  int j;
  int x;
  int xk;
  
  float piX;
  float piXk;
  float de;
  
  
  
  pt_stations depart=teteDepart;
  pt_stations arrivee=teteArrivee;
  /* On affiche les numeros de stations de depart et d'arrivee */
  printf("Station de depart :");
  AfficherListe(depart);
  printf("\nStation d'arrivee :");
  AfficherListe(arrivee);
  
  
  
  
  /* Initialisation de la liste de l'arborescence*/
  pt_listeChemins courant;
  courant=tete;
  
  /* Initialisation des portentiels a l'infini*/
  for (i=0;i<n;i++)
    {
      Graphe[i]->potentiel=INFINITY;
      Graphe[i]->marque=FAUX;
    }
  
  if(depart==NULL)
    {
      /* Le sommet de depart a mal ete entre */
      fprintf(stderr,"Erreur dans le choix des stations\n");
      exit (4);
    }
  
  /* On initialise les potentiels des racines a 0 et on ne les marque pas*/
  /* On cree donc une racine unique fictive */
  while(depart!= NULL)
    { 
      Graphe[depart->numStation]->potentiel=0.0;
      Graphe[depart->numStation]->marque=FAUX;
      Graphe[depart->numStation]->stationMere=depart->numStation;
      depart=depart->suivant;
    }
  
  
  depart=teteDepart;
  if(depart==NULL)
    {
      fprintf(stderr,"Erreur sur la liste de depart\n");
      exit (5);
    }
  
  /* On prend la premiere station de la liste de depart car toutes au minimum*/
  xk=depart->numStation;
  piXk=Graphe[xk]->potentiel;
  
  /* On marque ce sommet puis on l'ajoute a notre arborescence */
  Graphe[xk]->marque=VRAI;
  courant=ajouterFin(courant,depart->numStation,Graphe[depart->numStation]->potentiel,depart->numStation);
  
  
  /* numero d'etape */
  k=1;
  
  /* boucle parcourant le graphe : On s'arrete des que l'un des sommets et 
     marque et il appartient a la liste d'arrivee */
  while (!estDansListe(xk,arrivee) && k<n && piXk<INFINITY)
    {
      
      /* pour tout arcs (xk x) avec x n'appartenant pas a S*/ 
      
      if (Graphe[xk]->successeurs!=NULL){
	
	while (Graphe[xk]->successeurs!=NULL)
	  {
	    x=Graphe[xk]->successeurs->numStation;
	    
	    if(Graphe[x]->marque == FAUX)
	      {
		de=Graphe[xk]->successeurs->poids;
		piX=Graphe[x]->potentiel;
		
		
		if (piX > (piXk + de))
		  {
		    /* on change le potentiel */
		    Graphe[x]->potentiel=piXk+de;
		    
		    /* A(x)=e on ajoute l'arc a l'aborescence */
		    /* on met le sommet qui a marque x */
		    Graphe[x]->stationMere=xk;
		  }
	      }
	    Graphe[xk]->successeurs= Graphe[xk]->successeurs->suivant;
	  }
      }
            
      k++;  
      
      /* choisir x n'appartenant pas a S tel que potentiel de x soit minimum*/
      j=minimum(Graphe,n);
      Graphe[j]->marque=VRAI;
      
      xk=j;
      piXk=Graphe[xk]->potentiel;
      
      /* S recoit S union xk */
      courant=ajouterFin(courant,xk,Graphe[xk]->potentiel,Graphe[xk]->stationMere);
    }
  /* courant contient la fin de S donc on peut genere le chemin */
  genereChemin(courant);
}




/**************************************************************************
 * procedure construireFichierSolution(pt_Sommets Graphe[],int n)
 *
 * Donnee(s) : Graphe : tableau de Sommets representant le graphe
 *             n : nombre de sommets
 * Donnee(s)/Resultat(s) :
 * Resultat(s) :
 *
 * Cette procedure construit le fichier solution.graph de sortie contenant
 * tous les sommets et les arcs selectionnes dans notre liste de reponses 
 **************************************************************************/

void construireFichierSolution(pt_Sommets Graphe[],int n)
{
  int i;
  pt_listeChemins ptr;
  FILE *f2;
  
  /* ouverture du fichier en ecriture*/
  f2=fopen("solution.graph","w");
  if(f2==NULL)
    {
      fprintf(stderr,"Erreur dans l'ouverture du fichier en ecriture\n");
      exit (3);
    }
  
  /* premiere ligne contenant nombre de sommets et nombre d'arcs*/
  fprintf(f2,"%d %d\n",n,nbSommetsMarques);
  
  /* ajout du commentaire et des noms de sommets */
  fprintf(f2,"noms sommets\n");
  for (i=0;i<n;i++)
    {
      fprintf(f2,"%d %s\n", i, Graphe[i]->nomStation);
    }

  /* ajout du commentaire et des coordonnees des sommets */
  fprintf(f2,"coord sommets\n");
  for (i=0;i<n;i++)
    {
      fprintf(f2,"%d %d %d\n", i, Graphe[i]->coordX,Graphe[i]->coordY);
    }
  
 /* ajout du commentaire et des arcs a partir du chemin final genere */
  fprintf(f2,"arcs values\n");
  ptr=tete;
  if (ptr!=NULL)
    {
      while (ptr!=NULL)
	{
	  fprintf(f2,"%d %d %f\n", ptr->stationMere,ptr->numStation, ptr->potentiel);
	  ptr=ptr->suivant;
	}
    }
  
  fclose(f2);
  
  /* On lance la visualisation */
  system ("PCC_Visu/graphe2ps.exe solution.graph &");
  system ("display solution.graph.eps &");
  
}






/**************************************************************************
 * fonction main(int argc, char **argv) retourne 0 si tout s'est bien passe
 *
 * Donnee(s) : argc : entier nombre de parametres
 *             argv : tableau dechaine de caractere contenant les parametres
 *             du fichier
 *
 * Procedure principale qui lance le chargement du fichier, la saisie des
 * stations de depart et d'arrivee, le calcul du plus court chemin, 
 * l'affichage du resultat et la generation du fichier de sortie
 **************************************************************************/

int main(int argc, char **argv)
{
  /* Verification du nombre de parametres */
  if (argc!=2)
    {
      fprintf(stderr,"Erreur Mauvais nombre de parametres usage : %s nomfichier\n",argv[0]);
      exit (1);
    }
  
  
  /* ouverture du fichier passe en parametre en lecture pour le chargement*/
  FILE *f;
  f=fopen(argv[1],"r");
  if(f==NULL)
    {
      fprintf(stderr,"Erreur dans l'ouverture du fichier passe en parametre\n");
      exit (2);
    }
  
  
  
  /* lecture du nb de sommets et du nombre d'arcs */
  int n;
  int m;
  fscanf(f,"%d %d\n",&n,&m);
  
  
  /* Creation du graphe */
  pt_Sommets Graphe[n];
  
  /* on lance le remplissage du graphe */
  chargement_Fichier(Graphe,n,m,f);
  fclose(f);
  
  
  /* On entre les stations de depart et d'arrivee */
  printf("Station de depart : \n");
  rechercheStations(Graphe,n,DEP);
  
  printf("Station d'arrivee : \n");
  rechercheStations(Graphe,n,ARRIV);
  
  
  /* Calcul du plus court chemin */
  dijkstra(Graphe, n);
  
  /* On affiche le plus court chemin */
  afficherResultat(tete,Graphe);
  
  /* On construit le fichier de sortie */
  construireFichierSolution(Graphe,n);
  
  return 0;
}





















